// $Id: rmannoyRefreshService.js 109 2006-04-28 00:58:59Z das-svn $

// I'm convinced that Javascript prototype inheritance is cruel trick
// played upon developers.  As best as I can figure out, the prototype
// search only kicks in when accessing the methods outside of the object
// itself, so trying to get to this.hejaz where hejaz is in the prototype
// won't work.  To get around this, I'm going to pollute up the namespace
// and hope that this stupid language is better at making closures.  This
// object gets to be the prototype object.

var prototype =
   Components.classes["@mozilla.org/array;1"].
   createInstance(Components.interfaces.nsIMutableArray);

function RmannoyRefreshService() {
   // Using the constructors to set attributes of an empty object
   // lays bare the senseless differences between object initialization
   // syntaxes and makes one cry out against the injustices of a callous and
   // gray world.
   // I use the object literal syntax since it can conceivably be used in
   // more places, and that's what I'm used to.
   var object = {
      supportedInterfaces:
         [Components.interfaces.rmannoyRefreshService,
          Components.interfaces.nsIMutableArray,
          Components.interfaces.nsIArray,
          Components.interfaces.nsIClassInfo,
          Components.interfaces.nsIObserver,
          Components.interfaces.nsISupports],

      observers: new Array(),

      // nsIClassInfo
      classDescription: "rmannoy meta refresh exceptions data source",
      classID: Components.ID("{c0bd0d4d-b166-446d-8970-7aca0af3d1da}"),
      contractID: "@gophernet.org/rmannoy/refresh-service;1",
      flags: 0,
      implementationLangauge: Components.interfaces.nsIProgrammingLanguage.JAVASCRIPT,

      getHelperForLanguage: function(language) {
         if (language == Components.interfaces.nsIProgrammingLanguage.JAVASCRIPT) {
            return this;
         } else {
            return null;
         }
      },

      // Thanks for not explaining this one, XPCOM guys.
      // This is apparently how you handle "out" variables.  Pay attention.
      getInterfaces: function(count) {
         // remove nsISupports from the list
         var slice = this.supportedInterfaces.slice(0, this.supportedInterfaces.length-2);
         count.value = slice.length;
         return slice;
      },

      // rmannoyRefreshService
      initList: function(patterns) {
         var list = patterns.split(" ");

         prototype.clear();
         for (var key in list) {
            prototype.appendElement(this.boxPattern(list[key]), false);
         }

         for (var key in this.observers) {
            this.observers[key].observe(this, "rmannoy:changed", patterns);
         }
      },

      toString: function() {
         var str = "";
         var e = prototype.enumerate();

         // Set the first one without a space
         if (e.hasMoreElements()) {
            str = this.unboxPattern(e.getNext());
         }

         while (e.hasMoreElements()) {
            str = str + " " + this.unboxPattern(e.getNext());
         }

         return str;
      },

      isInList: function(uri) {
         // Thanks to adblock and greasemonkey for the idea.  It's not nearly
         // as difficult as greasemonkey makes it out to be, though.
         function RE(pattern)
         {
            // Convert the * wildcard into a .* regexp and escape everything
            // else regexpy.  Not replacing ? with ., since ? is a common URL
            // character and more useful to match literally.
            return new RegExp("^" +
               encodeURI(pattern).
               replace(/\./g, "\.").
               replace(/\*/g, ".*").
               replace(/([$^+?=!:|\\\/()\[\]{}])/g, "\\$1") +
               "$");
         }

         var e = prototype.enumerate();
         while (e.hasMoreElements()) {
            var pattern = this.unboxPattern(e.getNext());
            if (RE(pattern).test(uri)) {
               return true;
            }
         }
         return false;
      },

      addObserver: function(observer) {
         this.observers.push(observer);
      },

      removeObserver: function(observer) {
         for (var i in this.observers) {
            if (this.observers[i] == observer) {
               delete this.observers[i];
               return;
            }
         }
      },

      boxPattern: function(pattern) {
         var strObj = Components.classes["@mozilla.org/supports-string;1"].
            createInstance(Components.interfaces.nsISupportsString);

         strObj.data = pattern;
         return strObj;
      },

      unboxPattern: function(patternObj) {
         return patternObj.
            QueryInterface(Components.interfaces.nsISupportsString).
            data;
      },

      // nsIObserver
      // this object gets registered as an observer when added to the
      // category in registerSelf
      observe: function(subject, topic, data) {
         // app-startup happens before the user preferences are read, so I
         // can't read the preference yet, but I can add myself as an observer
         // for the preference
         var prefObserver = Components.classes["@mozilla.org/preferences-service;1"].
            getService(Components.interfaces.nsIPrefBranch2);
         switch (topic) {
            case "app-startup":
               // register this function as a preference observer
               prefObserver.addObserver("rmannoy.meta_refresh.exceptions",
                  this, false);
               
               break;
            case "nsPref:changed":
               if (data != "rmannoy.meta_refresh.exceptions") {
                  break;
               }

               var prefStr = Components.classes["@mozilla.org/preferences-service;1"].
                  getService(Components.interfaces.nsIPrefService).
                  getBranch("").getCharPref("rmannoy.meta_refresh.exceptions");

               this.initList(prefStr);
               break;
         }
      },

      // nsISupports
      QueryInterface: function(interfaceId) {
         for (var i in this.supportedInterfaces) {
            if (this.supportedInterfaces[i].equals(interfaceId)) {
               return this;
            }
         }
         throw Components.results.NS_ERROR_NO_INTERFACE;
      },

      // passthrough functions because either XPCOM or Javascript is a
      // pile of shit that can't handle any kind of sane inheritance
      appendElement: prototype.appendElement,
      clear: prototype.clear,
      insertElementAt: prototype.insertElementAt,
      removeElementAt: prototype.removeElementAt,
      replaceElementAt: prototype.replaceElementAt,

      enumerate: prototype.enumerate,
      indexOf: prototype.indexOf,
      queryElementAt: prototype.queryElementAt
   };

   return object;
};

RmannoyRefreshService.prototype = prototype;
// the O'Reilly javascript book recommends setting the constructor property
// of the prototype at this point, but this isn't possible with XPCOM objects,
// and the fact that such a thing might be needed makes me wonder what kind of
// God would allow such a language to exist.

var rmannoyRefreshService = new RmannoyRefreshService();

// Everything below this line is to shoehorn the above object through XPCOM

// create a nsIFactory implementation to return the service
var Factory = {
   // nsIFactory
   createInstance: function(outer, interfaceId) {
      // disable aggregation support.  Probably for the better, since
      // I don't even know what that is.
      if (outer != null) {
         throw Components.results.NS_ERROR_NO_AGGREGATION;
      }

      return rmannoyRefreshService;
   },

   // nsISupports
   QueryInterface: function(interfaceId) {
      if (!interfaceId.equals(Components.interfaces.nsIFactory) &&
          !interfaceId.equals(Components.interfaces.nsISupports)) {
         throw Components.results.NS_ERROR_NO_INTERFACE;
      }

      return this;
   }
};

// create a nsIModule implementation to use the factory
var Module = {
   // nsIModule
   canUnload: function(compMgr) { return true; },

   getClassObject: function(compMgr, classId, interfaceId) {
      if (!classId.equals(rmannoyRefreshService.classID)) {
         throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
      }

      return Factory.QueryInterface(interfaceId);
   },

   registerSelf: function(compMgr, location, loaderStr, type) {
      var catMgr = Components.classes["@mozilla.org/categorymanager;1"].
         getService(Components.interfaces.nsICategoryManager);

      // Make sure the component manager is behaving as a component manager
      compMgr = 
         compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

      // register with the component manager
      compMgr.registerFactoryLocation(rmannoyRefreshService.classID,
         rmannoyRefreshService.classDescription,
         rmannoyRefreshService.contractID,
         location,
         loaderStr,
         type);

      // I don't really know what this does, but all the cool kids are
      // doing it.
      catMgr.addCategoryEntry("app-startup",
         rmannoyRefreshService.classDescription,
         rmannoyRefreshService.contractID,
         true,
         true);
   },

   unregisterSelf: function(compMgr, location, loaderStr) {
      // do the opposite of what registerSelf did
      var catMgr = Components.classes["@mozilla.org/categorymanager;1"].
         getService(Components.interfaces.nsICategoryManager);

      compMgr = 
         compMgr.QueryInterface(Components.interface.nsIComponentRegistrar);

      compMgr.unregisterFactoryLocation(rmannoyRefreshService.classID, location);

      catMgr.deleteCategoryEntry("app-startup",
         rmannoyRefreshService.classDescription, true);
   },


   // nsISupports
   QueryInterface: function(interfaceId) {
      if (!interfaceId.equals(Components.interfaces.nsIModule) &&
          !interfaceId.equals(Components.interfaces.nsISupports)) {
         throw Components.results.NS_ERROR_NO_INTERFACE;
      }
   }
};

// This symbol needs to be exported to get all of the above to work
function NSGetModule(compMgr, fileSpec) {
   return Module;
}
